from datetime import date
from typing import Any

from PyQt6.QtCore import Qt, QDate
from PyQt6.QtWidgets import (
    QListWidgetItem,
    QCalendarWidget,
    QApplication,
    QWidget,
    QLineEdit,
    QTextEdit,
    QSpinBox,
    QCheckBox,
    QDoubleSpinBox,
)

from jbqt.common import consts
from jbqt.models import IChipsWidget


def get_item_value(item: QListWidgetItem) -> str:
    value = item.data(consts.LIST_ITEM_ROLE) or item.text()
    return value.strip()


def register_app(app: QApplication, icon_dir: str = "") -> None:
    consts.GlobalRefs.app = app
    consts.set_icon_dir(icon_dir)


def get_widget_value(widget: QWidget, key: str = "") -> Any:

    if isinstance(widget, QLineEdit):
        return widget.text()
    if isinstance(widget, QTextEdit):
        return widget.toPlainText().strip()
    if isinstance(widget, (QSpinBox, QDoubleSpinBox)):
        return widget.value()
    if isinstance(widget, IChipsWidget):
        return widget.values
    if isinstance(widget, QCheckBox):
        return widget.checkState == Qt.CheckState.Checked
    if isinstance(widget, QCalendarWidget):
        return widget.selectedDate().toString()

    print(f"No handler defined for {key} of type `{type(widget)}`")


def set_widget_value(widget: QWidget, value: Any) -> None:
    if isinstance(widget, (QLineEdit, QTextEdit)):
        widget.setText(str(value))
    elif isinstance(widget, QSpinBox):
        try:
            widget.setValue(int(value))
        except Exception as e:
            print(e)
            print(type(e))

    elif isinstance(widget, QDoubleSpinBox):
        try:
            widget.setValue(float(value))
        except Exception as e:
            print(e)
            print(type(e))

    elif isinstance(widget, IChipsWidget) and isinstance(value, list):
        widget.add_chips(value)
    elif isinstance(widget, QCheckBox):
        state = Qt.CheckState.Unchecked
        if isinstance(value, Qt.CheckState):
            state = value
        elif value is True:
            state = Qt.CheckState.Checked
        elif value is False:
            state = Qt.CheckState.Unchecked

        widget.setCheckState(state)

    elif isinstance(widget, QCalendarWidget):
        if isinstance(value, (date, QDate)):
            widget.selectedDate(value)



from re import Pattern

from PyQt6.QtCore import Qt, pyqtSignal, QObject, QEvent
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QScrollArea,
    QVBoxLayout,
    QWidget,
    QComboBox,
)

import jbutils as utils
from jbqt.common import consts
from jbqt.models import IChipsWidget
from jbqt.widgets.chip_button import ChipButton
from jbqt.widgets.multiselect import MultiSelectComboBox
from jbqt.widgets.widget_utils import debug_scroll_pos


class ChipsWidget(IChipsWidget):
    valuesChanged = pyqtSignal(list)

    def __init__(
        self,
        chips=None,
        data: dict = None,
        path: str = "",
        label: str = "",
        weight_re: Pattern = None,
        use_weight: bool = False,
        debug: bool = False,
    ):
        super().__init__()
        self.values = chips or []
        # Main layout
        main_layout = QVBoxLayout()
        self.debug = debug
        if label:
            main_layout.addWidget(QLabel(label))
        self.installEventFilter(self)

        self.use_weight = use_weight
        if use_weight:
            weight_re = weight_re or consts.TAG_RE

        self.weight_re = weight_re
        self.options = data
        self.data = data or {}
        self.data_path = path
        self.list_widget: QComboBox | MultiSelectComboBox = None

        if data is not None:
            if isinstance(data, dict):
                if path:
                    self.options = utils.get_nested(self.data, path, [])

                self.list_widget = MultiSelectComboBox(
                    data=self.options, selected=self.values
                )
                self.list_widget.selectedChanged.connect(
                    self.handle_multiselect_change
                )

            elif isinstance(data, list):
                self.list_widget = QComboBox()

                # self.list_widget.addItems(self.options)
                for option in self.options:
                    self.list_widget.addItem(option)
                init_value = self.values[0] if self.values else ""
                self.list_widget.setCurrentText(init_value)
                self.list_widget.currentTextChanged.connect(
                    self.handle_multiselect_change
                )
            self.valuesChanged.connect(self.update_multiselect)
            main_layout.addWidget(self.list_widget)

        # Scroll area for the chips
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)

        # Widget inside the scroll area to hold the buttons
        self.chip_widget = QWidget()
        self.chip_layout = QHBoxLayout(self.chip_widget)
        self.chip_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        # Add existing chips as buttons
        self.set_items(emit=False)

        scroll_area.setWidget(self.chip_widget)

        # Layout for the input and add button
        input_layout = QHBoxLayout()
        self.input_field = QLineEdit(self)
        self.add_button = QPushButton("Add", self)
        self.add_button.clicked.connect(self.add_chip)

        self.clear_button = QPushButton(consts.ICONS.TRASH("darkred"), "", self)
        self.clear_button.clicked.connect(self.remove_all)
        # self.clear_button.setStyleSheet("color: red")

        input_layout.addWidget(self.input_field)
        input_layout.addWidget(self.add_button)
        input_layout.addWidget(self.clear_button)

        main_layout.addWidget(scroll_area)
        main_layout.addLayout(input_layout)

        self.setLayout(main_layout)

    def eventFilter(self, obj: QObject, event: QKeyEvent) -> bool:
        focus = consts.GlobalRefs.app.focusWidget()

        if event.type() == QEvent.Type.KeyPress and focus == self.input_field:
            match event.key():
                case Qt.Key.Key_Return | Qt.Key.Key_Enter:
                    self.add_chip()
        return super().eventFilter(obj, event)

    def get_unweight_chip(self, tag: str) -> str:
        if self.weight_re and self.weight_re.search(tag):
            return self.weight_re.search(tag).groups()[0]
        return tag

    def get_chip_list(self, tags: list[str]) -> list[str]:
        return [self.get_unweight_chip(tag) for tag in tags]

    def same_values(self, values: list[str]) -> bool:
        return set(self.get_chip_list(values)) == set(
            self.get_chip_list(self.values)
        )

    def handle_multiselect_change(self, selections: list[str] | str) -> None:
        if isinstance(selections, str):
            selections = [selections]

        selections = utils.dedupe_list(selections)
        if self.same_values(selections):
            return

        # self.values = selections
        self.set_items(selections, debug=True)

    def update_multiselect(self, values: list[str]) -> None:
        if isinstance(self.list_widget, MultiSelectComboBox):
            self.list_widget.set_selected(values, emit=False)
        else:
            value = values[0] if values else ""
            self.list_widget.setCurrentText(value)

    def clear_widgets(self) -> None:
        while self.chip_layout.count():
            widget = self.chip_layout.takeAt(0).widget()
            if widget is not None:
                widget.deleteLater()

    def set_items(
        self, values: list[str] = None, emit: bool = True, debug: bool = False
    ) -> None:
        self.clear_widgets()
        # TODO: Temp fix for weighted/custom tags and incoming multiselect values
        if values:
            tags = self.get_tag_list(self.values)
            for value in values:
                if value not in tags:
                    self.values.append(value)

            self.values = utils.dedupe_list(self.values)
        self.values = [value for value in self.values if value]
        for item in self.values:
            if not consts.GlobalRefs.debug_set:
                debug = True
                consts.GlobalRefs.debug_set = True
            else:
                debug = False
            chip_button = ChipButton(
                item,
                self.update_chip,
                self.remove_chip,
                debug=debug,
                weight_re=self.weight_re,
                use_weight=self.use_weight,
            )
            """ chip_button.setToolTip(
                f"Tag Count: {consts.TAG_COUNTS.get(item, "N/A")}"
            ) """

            self.chip_layout.addWidget(chip_button)

        if emit:
            self.emit_changes()

    def update_chip(self, prev_text: str, new_text: str) -> None:
        idx = self.values.index(prev_text)
        if idx >= 0:
            self.values[idx] = new_text

    def emit_changes(self):
        self.valuesChanged.emit(self.values)

    def add_chip(self):
        text = self.input_field.text().strip()
        if text and text not in self.values:
            chip_button = ChipButton(text, self.update_chip, self.remove_chip)

            self.chip_layout.addWidget(chip_button)
            self.values.append(text)
            self.emit_changes()
            self.input_field.setText("")

    def remove_chip(self, button: ChipButton):
        self.chip_layout.removeWidget(button)
        button.deleteLater()  # Schedule button widget for deletion
        self.values.remove(button.text)
        self.emit_changes()

    def add_chips(self, items: list[str]) -> None:
        utils.update_list_values(self.values, items)
        self.set_items()

    def remove_chips(self, items: list[str]) -> None:
        utils.remove_list_values(self.values, items)
        self.set_items()

    def remove_all(self, *_) -> None:
        self.values = []
        self.clear_widgets()
        self.emit_changes()



from collections.abc import Callable
from typing import Optional

from PyQt6.QtCore import QSize
from PyQt6.QtGui import QIcon, QColor


IconGetter = Callable[[str | QColor, QSize | int | None], QIcon]



import os
import re


import jbutils

from PyQt6.QtWidgets import QApplication, QMainWindow, QScrollArea, QWidget
from PyQt6.QtCore import Qt, QSize
from PyQt6.QtGui import QIcon, QPixmap, QPainter, QColor

from jbqt.types import IconGetter

TAG_RE = re.compile(r"^\((.+)\)(\d+\.\d+)$")
LIST_ITEM_ROLE = Qt.ItemDataRole.UserRole - 1

STYLES = jbutils.STYLES
COLORS = jbutils.COLORS


class GlobalRefs:
    app: QApplication | None = None
    main_window: QMainWindow | None = None
    scroll_area: QScrollArea | None = None
    server_active: bool = False
    debug_set: bool = False
    seed_widget: QWidget | None = None


class QtPaths:
    icon_dir: str | None = None


THEME_ICONS: dict[str, QIcon] = {
    "save": QIcon.fromTheme("media-floppy-symbolic"),
    "new_file": QIcon.fromTheme("document-new-symbolic"),
    "open": QIcon.fromTheme("folder-open-symbolic"),
    "copy": QIcon.fromTheme("edit-paste-symbolic"),
    "inspect": QIcon.fromTheme("docviewer-app-symbolic"),
    "clone": QIcon.fromTheme("edit-copy-symbolic"),
    "plus": QIcon.fromTheme("list-add-symbolic"),
    "minus": QIcon.fromTheme("list-remove-symbolic"),
    "times": QIcon.fromTheme("cancel-operation-symbolic"),
    "refresh": QIcon.fromTheme("view-refresh-symbolic"),
    "reload": QIcon.fromTheme("update-symbolic"),
    "circle_check": QIcon.fromTheme("selection-checked"),
    "circle_x": QIcon.fromTheme("application-exit"),
    "code": QIcon.fromTheme("input-tablet-symbolic"),
    "font-selection-editor": QIcon.fromTheme("font-select-symbolic"),
    "trash": QIcon.fromTheme("trash-symbolic"),
    "edit_data": QIcon.fromTheme("document-edit-symbolic"),
    "preview": QIcon.fromTheme("view-layout-symbolic"),
    "sync": QIcon.fromTheme("mail-send-receive-symbolic"),
}
STD_ICONS: dict[str, QIcon] = {
    "save": QIcon.fromTheme(QIcon.ThemeIcon.DocumentSave),
    "new_file": QIcon.fromTheme(QIcon.ThemeIcon.DocumentNew),
    "open": QIcon.fromTheme(QIcon.ThemeIcon.FolderOpen),
    "copy": QIcon.fromTheme(QIcon.ThemeIcon.EditPaste),
    "inspect": QIcon.fromTheme(QIcon.ThemeIcon.DocumentPageSetup),
    "clone": QIcon.fromTheme(QIcon.ThemeIcon.EditCopy),
    "plus": QIcon.fromTheme(QIcon.ThemeIcon.ListAdd),
    "minus": QIcon.fromTheme(QIcon.ThemeIcon.DialogError),
    "times": QIcon.fromTheme(QIcon.ThemeIcon.ApplicationExit),
    "refresh": QIcon.fromTheme(QIcon.ThemeIcon.SystemReboot),
    "reload": QIcon.fromTheme(QIcon.ThemeIcon.MediaPlaylistRepeat),
    "circle_check": QIcon.fromTheme(QIcon.ThemeIcon.MediaPlaybackStart),
    "circle_x": QIcon.fromTheme(QIcon.ThemeIcon.ProcessStop),
    "code": QIcon.fromTheme(QIcon.ThemeIcon.Computer),
    "font-selection-editor": QIcon.fromTheme(QIcon.ThemeIcon.EditFind),
    "trash": QIcon.fromTheme(QIcon.ThemeIcon.EditDelete),
    "edit_data": QIcon.fromTheme(QIcon.ThemeIcon.InputTablet),
    "preview": QIcon.fromTheme(QIcon.ThemeIcon.ZoomIn),
    "sync": QIcon.fromTheme(QIcon.ThemeIcon.MediaPlaylistShuffle),
}


def set_icon_dir(path: str) -> None:
    QtPaths.icon_dir = path


class SIZES:
    ICON_XS = QSize(16, 16)
    ICON_SM = QSize(20, 20)
    ICON_MD = QSize(22, 22)
    ICON_LG = QSize(24, 24)
    ICON_XL = QSize(32, 32)
    ICON_XXL = QSize(48, 48)

    @classmethod
    def get(cls, key: str, default: QSize | None = None) -> QSize:
        if hasattr(cls, key):
            return getattr(cls, key)
        return default or cls.ICON_MD


def icon_dir(file_name: str) -> str:
    if QtPaths.icon_dir:
        return os.path.join(QtPaths.icon_dir, file_name)
    return file_name


def recolor_icon(
    image_path: str, color: str | QColor, scale: QSize | None = None
) -> QIcon:
    if not color:
        color = QColor("black")
    if isinstance(color, str):
        color = QColor(color)

    if scale is None:
        scale = SIZES.ICON_MD
    pixmap = QPixmap(image_path)
    icon_key = os.path.splitext(os.path.basename(image_path))[0]

    if pixmap.isNull():
        fallback = THEME_ICONS.get(icon_key, STD_ICONS.get(icon_key))
        if fallback:
            pixmap = fallback.pixmap()

    recolored_pixmap = QPixmap(pixmap.size())

    recolored_pixmap.fill(
        QColor("transparent")
    )  # Ensure the background is transparent
    painter = QPainter(recolored_pixmap)
    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)

    # Draw the original pixmap
    painter.drawPixmap(0, 0, pixmap)

    # Apply the desired color
    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
    painter.fillRect(pixmap.rect(), color)

    painter.end()
    return QIcon(recolored_pixmap.scaled(scale, Qt.AspectRatioMode.KeepAspectRatio))


def get_icon(file_name: str) -> IconGetter:
    def getter(color: str | QColor = "", size: QSize | int | None = None) -> QIcon:
        if isinstance(size, int):
            size = QSize(size, size)

        return recolor_icon(icon_dir(file_name), color, size)

    return getter


class ICONS:
    SAVE = get_icon("save.svg")
    NEW = get_icon("new_file.svg")
    OPEN = get_icon("open.png")
    COPY = get_icon("copy.png")
    INSPECT = get_icon("inspect.svg")
    CLONE = get_icon("clone.svg")
    PLUS = get_icon("plus.png")
    MINUS = get_icon("minus.png")
    TIMES = get_icon("times.svg")
    REFRESH = get_icon("refresh.svg")
    RELOAD = get_icon("reload.png")
    CIRCLE_CHECK = get_icon("circle_check.svg")
    CIRCLE_TIMES = get_icon("circle_x.svg")
    CODE = get_icon("code.png")
    EDIT = get_icon("font-selection-editor.png")
    TRASH = get_icon("trash.png")
    EDIT_DATA = get_icon("edit_data.svg")
    PREVIEW = get_icon("preview.png")
    SYNC = get_icon("sync.png")

    @classmethod
    def get(cls, name: str, default: str = "") -> IconGetter | str:
        if hasattr(cls, name):
            return getattr(cls, name)
        else:
            return default



from re import Pattern
from typing import Callable

from PyQt6.QtCore import Qt, QEvent, QObject
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QFrame,
    QWidget,
)

from jbqt.common import consts
from jbqt.common.consts import STYLES, COLORS
from jbqt.models import IChipButton
from jbqt.widgets.simple import ClickableLabel
from jbqt.dialogs import InputDialog
from jbqt.widgets.widget_utils import (
    debug_scroll_pos,
    preserve_scroll,
)


class ChipButton(IChipButton):
    def __init__(
        self,
        tag: str,
        on_update: Callable,
        on_remove: Callable,
        use_weight: bool = False,
        weight_re: Pattern = None,
        debug: bool = False,
    ):
        super().__init__()

        self._tag = tag
        self.weight: float = 1.0
        self.use_weight: bool = use_weight

        if use_weight:
            weight_re = weight_re or consts.TAG_RE

        if weight_re:
            tag_match = weight_re.search(tag)

            if tag_match:
                re_tag, weight = tag_match.groups()
                self._tag = re_tag
                self.weight = float(weight)

        self.on_update = on_update
        self.on_remove = on_remove
        self.installEventFilter(self)

        # Layout for the tag button
        # self.frame = QFrame(self, Qt.WindowType.ToolTip)
        self.frame = QFrame()
        frame_layout = QHBoxLayout(self.frame)
        frame_layout.setSpacing(5)
        frame_layout.setContentsMargins(10, 0, 10, 0)

        self.tag_label = ClickableLabel(tag)
        self.tag_label.setFixedHeight(24)
        self.tag_label.clicked.connect(self.toggle_edit_mode)

        self.weight_button: QPushButton = None

        if use_weight:
            self.weight_button = QPushButton(consts.ICONS.CODE(), "")
            self.weight_button.clicked.connect(self.apply_weight)
            self.weight_button.setFixedWidth(24)

        self.remove_button = QPushButton(consts.ICONS.TRASH(), "")
        self.remove_button.clicked.connect(lambda: self.on_remove(self))
        self.remove_button.setFixedWidth(24)
        self.remove_button.setStyleSheet(STYLES.BG_DARK_RED)

        self.edit_line = QLineEdit(self)
        self.edit_line.hide()

        self.confirm_btn = QPushButton(consts.ICONS.CIRCLE_CHECK(COLORS.WHITE), "")
        self.confirm_btn.clicked.connect(self.edit_tag)
        self.confirm_btn.setFixedWidth(24)
        self.confirm_btn.setStyleSheet(STYLES.BG_DARK_GREEN)
        self.confirm_btn.hide()

        self.cancel_btn = QPushButton(consts.ICONS.CIRCLE_TIMES(), "")
        self.cancel_btn.clicked.connect(self.cancel)
        self.cancel_btn.setFixedWidth(24)
        self.cancel_btn.setStyleSheet(STYLES.BG_DARK_RED)
        self.cancel_btn.hide()

        self.widgets = [
            self.tag_label,
            self.edit_line,
            self.confirm_btn,
            self.cancel_btn,
            self.weight_button,
            self.remove_button,
        ]
        self.input_widgets = [self.edit_line, self.confirm_btn, self.cancel_btn]
        for widget in self.widgets:
            if widget:
                frame_layout.addWidget(widget)

        # Main layout
        self.main_layout = QHBoxLayout(self)
        self.main_layout.addWidget(self.frame)
        self.main_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        # Apply styling to the frame
        self.frame.setStyleSheet(
            f"""
            QFrame {{
                border: 1px solid #AAA;
                border-radius: 10px;
                background-color: #343439;
                max-height: 36px;
            }}
            QLabel {{
                border: none;
            }}
            QPushButton {{
                {STYLES.BG_GRAY}   
            }}
        """
        )
        self.setLayout(self.main_layout)

    @property
    def text(self) -> str:
        if self.weight == 1:
            return self._tag
        return f"({self._tag}){self.weight:g}"

    @text.setter
    def text(self, value: str) -> None:
        self._tag = value

    def eventFilter(self, obj: QObject, event: QKeyEvent) -> bool:
        focus = consts.GlobalRefs.app.focusWidget()
        """ if event.type() == QEvent.Type.KeyPress:
            print("focus", focus) """
        if event.type() == QEvent.Type.KeyPress and focus == self.edit_line:
            match event.key():
                case Qt.Key.Key_Return | Qt.Key.Key_Enter:
                    self.edit_tag()
                case Qt.Key.Key_Escape:
                    self.cancel()
        return super().eventFilter(obj, event)

    def toggle_hidden(self, hide: bool = True) -> None:
        for widget in self.widgets:
            if widget:
                is_input = widget in self.input_widgets

                if hide == is_input:
                    widget.hide()
                else:
                    widget.show()

    def cancel(self) -> None:
        self.toggle_hidden()

    def toggle_edit_mode(self):
        self.edit_line.setText(self._tag)
        self.edit_line.setMinimumWidth(self.tag_label.width() + 10)
        self.toggle_hidden(False)

    def emit_update(self, text: str = None, weight: float = None) -> None:
        text = text or self._tag
        if weight is None:
            weight = self.weight
        if weight == "":
            self.on_remove(self)
            return
        prev_text = self.text
        self._tag = text
        self.weight = weight
        self.tag_label.setText(self.text)
        self.on_update(prev_text, self.text)

    def update_weight(self, weight: float) -> None:
        self.emit_update(weight=weight)

    @preserve_scroll
    def edit_tag(self, *_):
        text = self.edit_line.text()
        self.toggle_hidden()

        self.emit_update(text=text)

    def apply_weight(self):
        weight_dialog = InputDialog(
            parent=self,
            title="Tag Weight",
            msg_str=f"Enter weight for `{self.text}`:",
            input_type=float,
            minimum=0,
            maximum=2,
            singleStep=0.1,
            value=self.weight,
        )
        weight_dialog.connect(self.update_weight)
        weight_dialog.exec()
